class ExceptionPexpect(Exception):
    value: str
    def __init__(self, value: str) -> None: ...
    def get_trace(self) -> str: ...

class EOF(ExceptionPexpect): ...
class TIMEOUT(ExceptionPexpect): ...
